######################################################################
#:CQCC::IOBase 
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2001, 2008. All Rights Reserved.
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
#

=head1 NAME 

CQCC::IOBase - Base class for user and system interaction methods

=head1 CLASS DESCRIPTION

This class provides user interactors (dialog boxes, menus, etc)
and system interactors to manage file input and output. It acts
both as a functioning class that provides text interactors and as
a base class that more advanced interactor classes are built
upon.

This class is not accessed directly but rather through a set of
interface functions which hide the object implementation, and handle
initializing the IOBase object automatically if it hasn't already been
created.  Clients call IODebugPrint(), IOPathname(), IOPrompt(),
IOMessage(), IOFileRead(), and IOFileWrite() rather than the
corresponding methods.  This allows many different classes to use
IOBase without having to directly manage it's creation.

SUPPORT POLICY: See TriggerCQCC.pm for the Rational Support Policy. 

=head1 METHODS

=cut

#NOTES: 
# (1) Any calls to Message or DebugPrint that might occur before 
#     Initialization is complete use the Int* methods so that they use 
#     pre-initialization safe implementations.
# (2) Any method calls that might occur before Initialization
#     is complete use $self->method() calls instead of IOmethod functions
# (3) Methods that explicitly check value of $_IOstd or specific properties
#     may call IO* functions since they know initialization is complete.

######################################################################
#::Package declarations
######################################################################

package CQCC::IOBase;
use Exporter;
@ISA = ('Exporter');
@EXPORT = qw(&IODebugPrint &IOFileRead &IOFileWrite &IOMessage 
             &IOFileAppend $IO_FILE_WRITE $IO_FILE_APPEND
             &IOFileOperation $IO_FILE_COPY $IO_FILE_RENAME
             &IOListSelect &IOParseListString &IOPathname &IOPrompt 
             &IOPromptLogin &IOPromptYesNo &IOConvertString &IOInitialized
             &IOLockOperation &IOResourceManager &IOExitHandler 
             $IO_LOCKOP_NET $IO_LOCKOP_VERIFY $IO_LOCKOP_UNLOCK
             &IOClassFactory &IOGetCleanupHandler &IOSetCleanupHandler 
             &IOTraceBack &IOPrint &IOPrintStream &IOstd &IOListTreeSelect
             &IOQuotePathname &IOCheckCodePage);
             
use strict;
use CQCC::Parm;
use CQCC::MsgCat_English;
use CQCC::Timer;
use IO::File;
use File::Basename;

require Config;

sub Version { return '2.1' };

######################################################################
#::Class Globals 
######################################################################

my %debugLevel = ( ALWAYS => 0, 
                   BASIC => 1, 
                   DETAILED => 2,
                   VERY_DETAILED => 3
                 );

my $_IOstd = undef;  # Instance of IObase used by accessor functions

my ($_ExitObject,    # Object to be called on exit
    $_ExitMethod,    # Method to be called on exit
    %TempFiles,      # Temp files - set to 1 if defined, 2 if used
    $journalHandle,  # Flag indicating journal input file is open
   );

my $sep;             # Path separator for current platform
my %managedResources = ();  # Resources managed by IOResourceManager
my %opLocks = ();           # Operations locked by IOLockOperation

our $IO_FILE_WRITE = ">";    # File output modes used by FileWrite
our $IO_FILE_APPEND = ">>";

our $IO_FILE_COPY = "COPY";  # File operations for IOFileOperation
our $IO_FILE_RENAME = "RENAME";

our $IO_FLOCK_READ = "ReadShared";   # Operations for FileLock operation
our $IO_FLOCK_WRITE = "WriteExclusive";
our $IO_FLOCK_UNLOCK = "Unlock";

our $IO_LOCKOP_NET = "Net";
our $IO_LOCKOP_VERIFY = "Verify";
our $IO_LOCKOP_UNLOCK = "Unlock";

my %messageLevels = ("" => 0,       # Translates message codes to numbers
                     "INFO" => 1,
                     "WARN" => 2,
                     "ERROR" => 3,
                     "FATAL" => 4);

my $perl5008006 = ($] ge "5.008006" ? 1 : 0);

######################################################################
#::function ClassInit

=over 6

=item C<ClassInit()>

This class function is automatically invoked to define class 
parameters using CQCC::Parm objects. Current parameters include: 

=over 4

=item C<CQCC_DEBUG>

Controls level of debugging output reported thru the IODebugPrint
function. Currently three levels are supported: 0 - no output; 1 - 
basic output (high level operations); and 2 - detailed output. 

=item C<CQCC_JOURNAL_MODE>

Controls internal test journaling capability to handle user input. 
Values are 'OFF' (default), 'RECORD' (outputs user input as special
tags) and 'PLAYBACK' (expects to find tags in journal file input).

=item C<CQCC_JOURNAL_FILE>

File used for input when CQCC_JOURNAL_MODE is set to 'PLAYBACK'.

=item C<CQCC_LOG_OUTPUT>

Controls recording all warning, error and fatal messages to a log
file for convenience later during problem diagnosis. Values are
'OFF' which disables log file output; 'OVERWRITE' (default) which
overwrites the log file each trigger session; and 'APPEND', which
will add to the log file (only recommended during debugging). If
the value is OVERWRITE_THREADS or APPEND_THREADS then separate 
log files will be used for different series based threads. These
options should only be used when the integration is being debugged
when multiple threads are in use. 

=item C<CQCC_I18N_ENABLE>

Enable conversion of local codepage strings to and from UTF8 to
properly display and handle non-ASCII data. Set to "ON" by default,
disabled by setting to "OFF". 

=item C<CQCC_ASCII_CHECK_ENABLE>

When local codepage conversion is either disabled or not possible
in a given release, this flag enables checking that strings are 
ASCII only and displays an error when non-ASCII data is being
used. Set to "ON" by default, disabled by setting to "OFF". 

=item C<CQCC_CONVERT_ERROR_MAX>

This flag provides the maximum number of lines that should be 
displayed in an individual code page conversion error or 
ASCII check conversion error. Set to 100 by default. If not
defined, then no limit is imposed on conversion errors. 

=item C<IOBaseInterface>

Name of IOBase subclass to be instanced by IOClassFactory().

=item C<CQCC_CTRL_C>

Enables the control-C handler that tries to recognize when an end
user has hit control C to terminate the trigger, allowing the trigger
to report this as if the user had canceled the operation. Values are
'TRUE' (default) which enables the handler and 'FALSE' which disables
it. 

=item C<CQCC_CACHE_ROOT>

Establishes an optional central directory for cache files to be written
to. If set, the provided directory must exist, be accessible from the 
client or server host actually running the triggers, and be writeable by 
individual users to add new user-host subdirectories for new sessions.
If not set, information is cached to the users home directory on the
current machine they are using as in previous releases. 

=item C<CQCC_CACHE_MGMT>

Provides one or more internal parameters that modify cache management. 
Must be a string in the format "name=value,name2=value2".
Currently the following are defined: 

=over 4

CM_CLEAN_HRS - Minimum number of hours that pass between attempts
by the trigger to clean up stale cache and session files. Default is 4.
If 0, then every trigger will attempt cleanup. 

CM_STALE_HRS - Number of hours after which cache and session files 
are considered stale and appropriate to delete. Default is 8.

CM_CQSESSION_STALE_MINS - Number of minutes after which CQWeb sessions
are considered stale and appropriate to delete. Default is 30. Should 
be aligned with how long the CQWeb server considers sessions viable. 

CM_LOCK_BREAK_MINS - Number of minutes after which operation locks
can be broken by competing processes. Default is 5.

CM_LOCK_WARNING_MINS - Number of minutes to wait before letting the user
know that we are waiting for a contested lock. Default is 1. If greater
than CM_LOCK_BREAK_MINS the user will not be notified. 

CM_WIN_UNIX - Must be set to TRUE if the central cache is on a UNIX platform
but accessed by windows clients, i.e. SAMBA. Default is FALSE.

= back

=item C<CQCC_SESSION_ID>

If set, this is used to establish an independent caching subdirectory;
by default it is set to "main".

=item C<CQCC_LOG_ERRORS>

If set, enables writing out all error messages above the specified 
error level to a log file in the session caching directory. By default
this is set to ""; acceptable values include INFO, WARN, ERROR, and FATAL.
The internal parameters in use at the time of the error are also logged. 

=back 

=back 

=cut

######################################################################

sub ClassInit { 
    if (&ParmDefined("CQCC_DEBUG")) {
        return;
    }     

    # Environment variable source only unless others added
    &SetParmDefaults(PCLASS=>"CQCC::IOBase", SRC_EV=>1);

    &DefineParm(NAME=>"CQCC_DEBUG", DEFAULT=>0, SRC_CFG=>1,
               RANGE_MIN=>0, RANGE_MAX=>3);

    &DefineParm(NAME=>"CQCC_JOURNAL_MODE", DEFAULT=>"OFF",
                RANGE_LIST=>["OFF", "RECORD", "PLAYBACK"]);
                
    &DefineParm(NAME=>"CQCC_JOURNAL_FILE", DEFAULT=>"");

    &DefineParm(NAME=>"CQCC_LOG_OUTPUT", DEFAULT=>"OFF",
                RANGE_LIST=>["OFF", "OVERWRITE", "APPEND", 
                             "OVERWRITE_THREADS", "APPEND_THREADS"],
                SRC_CFG=>1);

    &DefineParm(NAME=>"CQCC_LOG_ERRORS", SRC_CFG=>1, DEFAULT=>"");

    &DefineParm(NAME=>"CQCC_SESSION_ID", DEFAULT=>"main");

    # Config source only unless otherwise others added
    &SetParmDefaults(PCLASS=>"CQCC::IOBase", SRC_CFG=>1);

    &DefineParm(NAME=>"CQCC_I18N_ENABLE", DEFAULT=>"ON",
                RANGE_LIST => ["ON", "OFF"]);

    &DefineParm(NAME=>"CQCC_ASCII_CHECK_ENABLE", DEFAULT=>"ON",
                RANGE_LIST => ["ON", "OFF"]);

    &DefineParm(NAME=>"CQCC_CONVERT_ERROR_MAX", DEFAULT=>"50");

    &DefineParm(NAME=>"IOBaseInterface", 
                DEFAULT=>($ENV{ATRIA_FORCE_GUI} ?
                          "CQCC::IOClearPrompt" : "CQCC::IOBase"));

    &DefineParm(NAME=>"CQCC_CTRL_C", DEFAULT=>"TRUE",
                RANGE_LIST => ["FALSE", "TRUE"]);

    &DefineParm(NAME=>"CQCC_CACHE_ROOT", DEFAULT=>"");

    &DefineParm(NAME=>"CQCC_CACHE_MGMT", DEFAULT=>"");

    &SetParmDefaults(PCLASS=>"CQCC::IOBase", SRC_INT=>1);
    &DefineParm(NAME=>"CM_CLEAN_HRS", DEFAULT=>4,
                RANGE_MIN=>0, RANGE_MAX=>100);
    &DefineParm(NAME=>"CM_STALE_HRS", DEFAULT=>8,
                RANGE_MIN=>0.1, RANGE_MAX=>100);
    &DefineParm(NAME=>"CM_CQSESSION_STALE_MINS", DEFAULT=>30,
                RANGE_MIN=>5, RANGE_MAX=>500);
    &DefineParm(NAME=>"CM_LOCK_BREAK_MINS", DEFAULT=>5,
                RANGE_MIN=>1, RANGE_MAX=>100);
    &DefineParm(NAME=>"CM_LOCK_WARNING_MINS", DEFAULT=>1,
                RANGE_MIN=>0.25, RANGE_MAX=>100);
    &DefineParm(NAME=>"CM_WIN_UNIX", DEFAULT=>"FALSE",
                RANGE_LIST => ["TRUE", "FALSE"]);


    # GUI button text for Confirm and Cancel options
    &DefineParm(NAME=>"YES_TEXT", DEFAULT=>"");
    &DefineParm(NAME=>"NO_TEXT", DEFAULT=>"");

    &SetParmDefaults();
}

&ClassInit();

######################################################################
#::function ClassTest

=over 6

=item C<ClassTest(@args)>

This class function uses sets up a local environment for testing;
automatically creates an instance of the selected IOBase subclass; 
and invokes the object's SelfTest method for testing. 

=back

=cut

######################################################################

sub ClassTest {
    my @args = @_;
    my $obj = &IOstd();
    return $obj->SelfTest(@args);
}

######################################################################
#::method Initialize

=over 6

=item C<Initialize()>

This method initializes the object attributes used in translating
pathname prefixes, determining which operating system/platform is
in use, initializing message catalogs and managing general status
information. Implicitly called by IOStd when creating a new IOBase 
subclass object. 

=back

=cut

######################################################################

sub Initialize {
    my $self = shift;

    ##################################################
#::: Common Setup
    $self->{WINDOWS} = ($Config::Config{osname} eq "MSWin32") ? 1 : 0;
    $sep = $self->{WINDOWS} ? "\\" : "/";
    $self->{MSGCAT}= &IOClassFactory("MessageCatalogInterface");

    # L10N: By default assume message cat should output local code page
    $self->{MSGCAT}->Encoding("IO_CODEPAGE",$self->{IO_CODEPAGE});
    
    # ConvertString can't reference parms without hitting recursion
    # so at this point check CQCC_I18N_ENABLE and CQCC_ASCII_CHECK_ENABLE
    # and tell the MsgCat how to behave

    $self->{MSGCAT}->EnableConversion(
        (&GetParm("CQCC_I18N_ENABLE") eq "ON") ? 1 : 0,
        (&GetParm("CQCC_ASCII_CHECK_ENABLE") eq "ON") ? 1 : 0,
        &GetParm("CQCC_CONVERT_ERROR_MAX"));

    # Pass the IOConvertString and IOMessage function ptrs to parm.pm, 
    # since it is not available yet for parm. MsgCat will also borrow
    # this from Parm. 

    CQCC::Parm::SetConvertHandler(\&IOConvertString);
    CQCC::Parm::SetMessageHandler(\&IOMessage);

    $self->{JOURNAL_MODE} = &GetParm("CQCC_JOURNAL_MODE"); 

    ######################################################################
#::: Initialize temp directory - must end without file separate at end
    # If temp directory exists and is writeable use, else 

    my $tmpdir;

    if ($self->{WINDOWS}) {
        $tmpdir = $ENV{'TMPDIR'} ? $ENV{'TMPDIR'} :
                  ($ENV{'TMP'} ? $ENV{'TMP'} :
                  ($ENV{'TEMP'} ? $ENV{'TEMP'} :
                  (-w "c:\\TEMP" ? "c:\\TEMP" :                         
                  "")));
    } else {
        $tmpdir = $ENV{'TMPDIR'} ? $ENV{'TMPDIR'} :
                  (-w "/tmp" ? "/tmp" : ".");
    }
    if ($tmpdir eq "" || ! -d $tmpdir) {
        # Display a failure message and terminate the process
        # Override message queuing since IOClearPrompt needs temp dir
        delete $ENV{"ATRIA_WEB_GUI"};
        $self->IntMessage("FATAL", ["IO_E_INVALID_TEMP_DIR"]);
    }
    $self->{TEMP} = $tmpdir;

    ######################################################################
#::: Determine series key to make uniquely named caching and resource files

    # Note: Don't depend on CC being initialized by this point
    my $key = $ENV{"CLEARCASE_SERIES_ID"};
    my $testKey = $key;
    $testKey =~ s/[0|.|:]//g;
    
    # Case where there is no series id is a command line initiated 
    # operation calling cqcc_launch directly. Create a unique key
    # to avoid conflicts in the case where there are two command lines
    # running at once, as in automated testing. 
    # 
    # Case where Series id is all 0's means it wasn't initialized
    # so use Clearcase Process ID instead to get thread unique name

    if ($key eq "") {
        $self->{SERIES_KEY} = "NO_SERIES.$$";

    } elsif ($testKey eq "") {
        my $ccPid = $ENV{"CLEARCASE_PPID"};
        if ($ccPid && $ccPid ne "") {
            $self->{SERIES_KEY} = $ccPid;
        } else {
            $self->IntMessage("FATAL", ["IO_E_INVALID_SERIES_KEY"]);
        }

    # Case where Series id is not all 0's, make it useable as file suffix
    } elsif ($key) {
        $key =~ s/\./_/g;
        $key =~ s/\:/_/g;
        $self->{SERIES_KEY} = $key;
    }

    ######################################################################
#::: Determine host and user name

    # TODO: Hostname could be a subroutine that provides info on demand 
    # and caches instead of hardcoded here. Save time if not needed. 
    # TODO: In CC7.0/CCRC look for remote client instead of local host
    # need to understand needs between CQWeb* and IOBase parm names
    # TODO/RISK: What is "CLEARCASE_USER" as opposed to local user ? 

    my ($user, $host, $devNull);
    if ($self->{WINDOWS}) {    
        $user = $ENV{USERNAME};
        $devNull = "NUL";
    } else {
        $user = $ENV{LOGNAME} ? $ENV{LOGNAME} : $ENV{USER};
        $host = `uname -n`;
        chop($host);
        $devNull = "/dev/null";
    }
    # TODO: In server environment, use CLIENT host instead of local host
    # for parameter directory

    # Needs to be based on separate EV provided by ccrc (or apache?)
    my ($hostname) = gethostbyname($host);

    $self->{LOCALHOST} = $hostname;  # Local hostname that gethostbyname uses
    $self->{LOCALUSER} = $ENV{CLEARCASE_USER} ? $ENV{CLEARCASE_USER} : $user;
    $self->{DEVNULL} = $devNull;
    
    ######################################################################
#::: Determine ATRIA home
    # TODO: Still USED by anything ? Hard to use integration if cc not in path

    if ($self->{WINDOWS}) {
        if ($ENV{'ATRIAHOME'}) { 
            $self->{ATRIA} = $ENV{'ATRIAHOME'} . $sep . "bin";
            # Push the atria bin onto path if not already there
            $ENV{PATH} = $ENV{PATH} . ";$self->{ATRIA}";
        } else {
            my @pathList = split(";",$ENV{"PATH"});
            for (@pathList) {
                if (-e $_ . $sep . "cleartool.exe") {
                    $self->{ATRIA} = $_;
                    last;
                }
            }
        }
        $self->{HELPAPP} = "start";  # Starts independent subprocess
    } else {
        $self->{ATRIA} = $ENV{'ATRIAHOME'} ? $ENV{'ATRIAHOME'} . "/bin" : 
                                            '/opt/rational/clearcase/bin';
        # Push the atria bin onto path if not already there
        $ENV{PATH} = $ENV{PATH} . ":$self->{ATRIA}";
        $self->{HELPAPP} = "/opt/rational/clearcase/etc/display_url.sh";
    }
    if (! defined $self->{ATRIA}) { 
        $self->IntMessage("FATAL", ["IO_E_INVALID_ATRIA_DIR"]);
    }

    ######################################################################
#::: Initialize parameter directory name(s)

    my $centralDir = &GetParm("CQCC_CACHE_ROOT");
    if (! $ENV{"ATRIA_WEB_GUI"}) {
       if ($self->{WINDOWS}) {
            if (defined $ENV{APPDATA}) { 
                $self->{PARM_USER} =
                    $self->make_dir($ENV{APPDATA},"Rational","CQCC");
            } elsif ($centralDir eq "") {
                $self->IntMessage("FATAL", [ "IO_E_CCRC_REQTS"]);
            }
        } else {
            my $user = $ENV{CLEARCASE_USER} ? $ENV{CLEARCASE_USER} : 
                       ( $ENV{LOGNAME} ? $ENV{LOGNAME} : $ENV{USER});
            my ($name,$pass,$uid,$gid,$quota,$comment,$gcos,$dir,$shell) =
                getpwnam($user);
            $self->{PARM_USER} = $dir;
        }
    }
    $self->{PARM} = $self->{PARM_USER};  # Default location

    ######################################################################
#::: Setup centralized per user/host/session subdirectory

    if ($ENV{"ATRIA_WEB_GUI"}) {
        my $CQCCext701 = 0;
        eval { 
            require CQCCext;
            if (defined &CQCCext::encrypt_password) {
                $CQCCext701 = 1;
            }
        };
        if (($CQCCext701 == 0) || ($centralDir eq "")) {
            $self->IntMessage("FATAL", [ "IO_E_CCRC_REQTS"]);
        }
    }
    if ($centralDir) { 
        if ($ENV{"CCRC_SESSION_ID"}) {
            &ForceParm("CQCC_SESSION_ID",$ENV{"CCRC_SESSION_ID"});
        }
        my $sessionID = "session." . &GetParm("CQCC_SESSION_ID");
        $self->{PARM} = $self->make_dir($centralDir, 
            $self->{LOCALUSER}, $self->{LOCALHOST}, $sessionID);
    }

    my $cacheMgmt = &GetParm("CQCC_CACHE_MGMT");
    if (defined $cacheMgmt) {
        #Remove white space and newlines to facilitate easier formatting
        $cacheMgmt =~ s/\s//g;
        $cacheMgmt =~ s/\n//g;
        my @parmPairs = split(",",$cacheMgmt);
        for (@parmPairs) {
            my ($name,$value) = split("=",$_);
            &ForceParm($name,$value);
        }
    }

    ######################################################################
#::: Determine commonly used paths once all the pathname dirs known

    # PARM_SERIES is same directory as PARM but adds series suffix
    $self->{PARM_SERIES} = $self->{PARM}; 
    # In server environment, there is no user home directory to use
    if ($ENV{"ATRIA_WEB_GUI"}) {
        $self->{PARM_USER} = $self->{PARM};
    }

    # Set LOG_FILE to PARM or PARM_SERIES based on debug option
    # By this point directory log file goes into should exist
    my $logDir = (&GetParm("CQCC_LOG_OUTPUT") =~ /THREADS/) ? "PARM_SERIES" : "PARM";
    my $prefix = $ENV{"CQCC_LOG_PREFIX"} ? $ENV{"CQCC_LOG_PREFIX"} : "";
    $self->{LOG_FILE} = $self->Pathname($logDir,$prefix . "cqcc_output.log");
    $self->{ERR_FILE} = $self->Pathname("PARM",$prefix . "cqcc_error.log");

    my $openFlag = (&GetParm("CQCC_LOG_OUTPUT") =~ /OVERWRITE/)
                    ? $IO_FILE_WRITE : $IO_FILE_APPEND;
    my $time = localtime(time);
    $self->PrintLog($openFlag, "\n::============================ ($time)\n" . 
                    "ClearCase/ClearQuest log\n\n" . 
                    # ATE references are for automated testing and will show nothing
                    # outside a test harness environment
                    ($ENV{"CQCC_ATE_TEST_CASE"} ? 
                        "ATE TestCase: " . $ENV{"CQCC_ATE_TEST_CASE"} . "\n" : "") .
                    ($ENV{"CQCC_ATE_TEST_PURPOSE"} ? 
                        "ATE TestPurpose: " . $ENV{"CQCC_ATE_TEST_PURPOSE"} . "\n" : "") .
                    "Series: $self->{SERIES_KEY} CC: $ENV{CLEARCASE_PPID} " .
                        "OP: $ENV{CLEARCASE_OP_KIND} THREAD: $ENV{CQCC_SUB_THREAD}\n"
                    );

    # Initialization debug statements go here, after initialization
    $self->DebugPrint("DETAILED", 
                      "IOBase::Initialize \n" .
                      "  localhost: $self->{LOCALHOST} " . 
                      " user: $self->{LOCALUSER} devnull: $self->{DEVNULL} \n" . 
                      "  Series Key: $self->{SERIES_KEY} \n" . 
                      "  PARM: $self->{PARM} \n" . 
                      "  PARM_USER: $self->{PARM_USER} \n" . 
                      "  ATRIA: $self->{ATRIAHOME} \n" . 
                      "  HELPAPP: $self->{HELPAPP} \n" . 
                      "  TEMP: $self->{TEMP}\n");


    # Set YES_TEXT and NO_TEXT to ENTER/RETURN for cmd-line interface
    if ($Config::Config{osname} eq "MSWin32") {
        &ForceParm("YES_TEXT", $self->{MSGCAT}->Message(["BUTTON_TEXT_ENTER"]));
        &ForceParm("NO_TEXT", $self->{MSGCAT}->Message(["BUTTON_TEXT_ENTER"]));
    }
    else {
        &ForceParm("YES_TEXT", $self->{MSGCAT}->Message(["BUTTON_TEXT_RETURN"]));
        &ForceParm("NO_TEXT", $self->{MSGCAT}->Message( ["BUTTON_TEXT_RETURN"]));
    }

    $self->{INITIALIZED} = 1;
}

######################################################################
#::IOInitialized
######################################################################

sub IOInitialized {
    return (defined $_IOstd && $_IOstd->{INITIALIZED} eq 1) ? 1 : 0;
}

######################################################################
#::make_dir 
# 
# Utility function to make new subdirectories if needed 
# base_dir is required directory that must already exist
# @subdirList is the hierarchy of subdirectories under base_dir
# that should be created if they don't already exist. 
# 
# Example - $io->make_dir("base", "b", "c")
# will create base/b/c if base already exists. 
#
# Output: list
#     Result - 1 if successful, 0 if not
#     string - pathname if successful, else error code
######################################################################

sub make_dir {
    my ($self,$base_dir,@subdirList) = @_;
    # TODO: Consider a IODirOperation method for mk/copy/del/readdir

    if (! -d $base_dir) {
        $self->IntMessage("FATAL", [ "IO_E_INVALID_ROOT_DIR", $base_dir]);
        # Return in case message queueing is enabled
        return;
    }

    my $subdir = $base_dir;
    for (@subdirList) {
        $subdir .= $sep . $_;
        if (! -d  $subdir) {
            # On windows, try to create directory with Clearcase group access
            if ($self->{WINDOWS} && &GetParm("CM_WIN_UNIX") eq "FALSE") {
                eval { 
                    require CQCCext;
                    if (defined &CQCCext::make_cc_dir) {
                        CQCCext::make_cc_dir($subdir);
                    }
                }
            };
            if (! -d $subdir) {
                mkdir $subdir, 0700;
            }
            if (! -d $subdir ) {
                $self->IntMessage("FATAL", 
                    [ "IO_E_FAILED_MKDIR", $subdir, $!]);
                return;
            }
        }
    }
    return $subdir;
}

######################################################################
#::function IOstd
# This internal function is called to get access to the current
# IOBase object, creating it if necessary. 
######################################################################

sub IOstd {
    if (! defined $_IOstd) {
        $_IOstd = &IOClassFactory("IOBaseInterface");
        $_IOstd->Initialize();

        # Message Queuing - Error messages seen before initialization
        # is complete in a CCRC environment won't be visible to the 
        # user unless they are delayed until IOClearPrompt is able to 
        # display them post-initialization.

        my $msg_queue = $_IOstd->{MSG_QUEUE};
        if (defined $msg_queue) {
            my @msgQueue = @{ $msg_queue };
            for (@msgQueue) {
                my (@args) = @{ $_ };
                &IOMessage(@args);
            }
        }
    }
    return $_IOstd;
}

######################################################################
#::function IOCompleteSetup
# This internal function is called once during main trigger startup
# to create a new IOBase class with the correct class and message
# catalog. 
######################################################################

sub IOCompleteSetup {
    if (defined $_IOstd) {
        $_IOstd->Shutdown();   # Shutdown existing object 
        $_IOstd = undef;
    }
    IOstd(); # Create new object
}

######################################################################
#::function IOCheckCodePage
# Windows only check for OEM code page problems

sub IOCheckCodePage {
    if ((&GetParm("CQCC_I18N_ENABLE") ne "OFF") && $perl5008006 &&
        ($Config::Config{osname} eq "MSWin32")) {

        # The client code page is the OEM code page for the operating system on
        # which BCP is run. Microsoft Windows NT, Microsoft Windows 95, and
        # Windows 98 operating systems have both an ANSI code page and an OEM
        # code page; these are determined by the country settings. The
        # computer's ANSI code page is stored in the following registry value:
        # 
        # HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Nls\CodePage\ACP
        # 
        # For U.S. English versions of Windows NT, Windows 95, and Windows 98,
        # the default OEM code page is 437 (U.S. English). The computer's OEM
        # code page is stored in the following registry value:
        # 
        # HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Nls\CodePage\OEMCP
        # 

        eval {
            require Win32::Registry;
            my $CodePageKey = "SYSTEM\\CurrentControlSet\\Control\\Nls\\CodePage";
            my $hkey;
            my %values;

            $::HKEY_LOCAL_MACHINE->Open($CodePageKey,$hkey) || die $!;
            $hkey->GetValues(\%values);
            $hkey->Close();
            my $acp = $values{"ACP"}->[2];
            my $chcpCmd = `chcp`;
            if ($chcpCmd =~ /\:\s*([0-9]*)/) {
                my $chcp = $1;
                if ($acp =~ /cp([0-9]*)/) {
                    $acp = $1;
                }
                if ($chcp ne $acp) {
                    &IOMessage("WARN", 
                        ["IO_E_OEM_ACP_MISMATCH", "cp$acp", "cp$1", $acp]);
                }
            }
        };
        if ($@) { 
            &IODebugPrint("DETAILED",
                "Unable to determine ActiveCodePage from Windows Registry");
        }
    }
}

######################################################################
#::function IOClassFactory

=over 6

=item C<IOClassFactory(classInterface)>

This function supports the ClassFactory methods in all classes
by trying to make an instance of the classInterface, confirm that a
valid object is created and report errors if necessary. 

=back

=cut

######################################################################

sub IOClassFactory {
   my ($classInterface, $classArgs) = @_;

   my $className = &GetParm($classInterface);
   my $t1 = &TimerStart("IOBase::IOClassFactory $className\n");
   eval "require $className";
   my $obj = eval "new $className($classArgs)";
   if (! defined $obj) {
       if (&IOInitialized()) { 
           &IOMessage("FATAL", 
              [ "IO_E_FACTORY_FAILED", $classInterface, $className ]);
       } else {
           # Should only be run during MsgCat or IOBase creation
           print"Class_Factory_Failed making $classInterface " . 
               "($className)\n";
           exit(2);
       }
   } 
   &TimerStop($t1);
   return $obj;
}

######################################################################
#::function IOConvertString

=over 6

=item C<IOConvertString($src_string_or_array, $src_encoding, $dst_encoding)>

Takes a string or string array reference in the given src_encoding and 
converts it to the dst_encoding using MsgCat::ConvertString.

=back

=cut

######################################################################

sub IOConvertString {
    if (&IOstd->{MSGCAT}) {
        return &IOstd->{MSGCAT}->ConvertString(@_);
    }
    else {
        &IOMessage("FATAL", [ "IO_E_MSGCAT_NOT_INITIALIZED" ]);
        return undef;
    }
}

######################################################################
#::function IODebugPrint

=over 6

=item C<IODebugPrint($type, $message)>

This accessor function optionally prints out debugging information.
Its output is controlled by the value of CQCC_DEBUG class parameter
which enables zero or more types to be enabled for printing: if 0, no
debug information is displayed; if 1, then Basic information is shown;
if 2, then detailed information is shown.

The type parameter is a string in the list "ALWAYS", "BASIC", and 
"DETAILED". The intention is to allow multiple "types" that may later
be split out into more debug levels if warranted. 

The message parameter is currently a single string not a message
catalog tag set. 

=back

=cut

######################################################################

sub IODebugPrint {
    return &IOstd->DebugPrint(@_);
}

######################################################################
#::method IntDebugPrint - Internal DebugPrint routine
# Internal version of DebugPrint that will only run if 
# the default IO object was successfully initialized
######################################################################

sub IntDebugPrint {
    my ($self,@args) = @_;
    if (&IOInitialized()) { 
        return &IODebugPrint(@args);
    }
}

######################################################################
#::method DebugPrint
######################################################################

sub DebugPrint {
    my ($self, $type, $message) = @_;  

    # Type compared to debug flag to filter output

    my $level = $debugLevel{$type};
    my $ev = &GetParm("CQCC_DEBUG");

    if (defined $level && $ev >= $level)
    {
       if (defined $ENV{"CQCC_UTIL_DEBUG"}) {
           $self->PrintLog($IO_FILE_APPEND,$message,"\n");
       } else {
           my ($status,$msgLocal) = 
               &IOConvertString($message,"UTF8_CODEPAGE","LOCAL_CODEPAGE"); 
           $self->Print($msgLocal,"\n");
       }
    }
}

######################################################################
#::function IOExitHandler

=over 6

=item C<IOExitHandler(status)>

This accessor function handles fatal errors, aborts, and normal exits
from the application. If a cleanup handler is defined that is called
before the program exits.

Status codes include the following: 
   0 = successful operation
   1 = abort/canceled operation
   2 = fatal error

=back

=cut

######################################################################

sub IOExitHandler {
    return &IOstd->ExitHandler(@_);
}

######################################################################
#::method ExitHandler
######################################################################

sub ExitHandler {
    my ($self, $status) = @_;
    my $exitObject = $_ExitObject;
    my $exitMethod = $_ExitMethod;    
    my $pathname;

    &IOSetCleanupHandler();  # Clear existing cleanup handler

    # Run the cleanup handler only on abort or fail exit status
    if ($status && $exitMethod) { 
        my $cmd;
        if ($exitObject) {
            $cmd = "\$exitObject->$exitMethod";
        } else {
            $cmd = "$exitMethod";
        }
        $self->IntDebugPrint("BASIC",
            "IOBase::ExitHandler - calling cleanup handler: $cmd\n");
        eval $cmd;
    } 

    # Release any managed resources for reused by other processes
    $self->ResourceManager(undef,0,0);
    $self->LockOperation(undef,$IO_LOCKOP_UNLOCK);

    &TimerLog();

    # Delete all the tempfiles we created before exiting
    for $pathname (keys %TempFiles) {
        if ($TempFiles{$pathname} == 2) { 
            unlink $pathname;
        }
    }

    $self->Shutdown($status);

    $self->IntDebugPrint("BASIC",
        "IOBase::ExitHandler - exit status: $status\n");

    # Cover over a CQPerl return code problem
    # Don't use IOPath because PC doesn't use PIDs

    my $retFile =
        $self->{TEMP} . $sep . "cqcc_launch.$ENV{CLEARCASE_PPID}";

    $self->FileWrite($retFile, $status);
    exit $status;
}

######################################################################
#::function IOFileAppend

=over 6

=item C<IOFileAppend(filename, item(s))>

This accessor function will open filename in append mode, write the 
items list to it, and close the file. Each value is written out as
a single line in the file. While not a broad purpose file writer, it
is intended to work in tandem with FileRead which reads one item to a
line.

=back 

=cut

######################################################################

sub IOFileAppend {
    my ($filename, @args) = @_;
    return &IOstd->FileOutput($filename, $IO_FILE_APPEND, @args);
}

######################################################################
#::function IOFileRead

=over 6

=item C<IOFileRead(filename, delete-flag)>

This accessor function will open, read from and close the requested
filename, returning its contents in a list of scalar values. It will
optionally delete it when complete if delete-flag is 1.

(item1,item2,...) = $obj->FileRead(filename, delete-flag);

Each value being returned is a single line in the file. While not 
a broad purpose file reader, it is intended to work in tandem with
FileWrite which writes one item to a line.

=back

=cut

######################################################################

sub IOFileRead {
    return &IOstd->FileRead(@_);
}

######################################################################
#::method FileRead
######################################################################

sub FileRead { 
    my ($self, $filename, $delete) = @_;

    my @answerList = ();
    my $answer;
        
    # Strip off enclosing quotes
    if (substr($filename,0,1) eq '"') {
        $filename = substr($filename,1,length($filename)-2);
    } 

    if (open (INFILE, $filename)) {
        $self->FileLock($IO_FLOCK_READ, "IOBase::FileRead", \*INFILE, $filename);

        while ($answer = <INFILE>) {
            chomp $answer;
            $answer =~ s/\r$//;   # If cross-platform remove extra line terminator
            push @answerList, $answer;
        }
        $self->FileLock($IO_FLOCK_UNLOCK, "IOBase::FileRead", \*INFILE, $filename);
        if ($delete) {
            unlink $filename;
        }
        return @answerList;
    }
    return;
}

######################################################################
#::method FileLock 
# Internal method to lock/release file locks
# 
# operation is $IO_FLOCK_WRITE(lock-exclusive), $IO_FLOCK_READ(lock-shared), 
# or $IO_FLOCK_UNLOCK(release lock). 
# 
#   Note that closing a file automatically unlocks it 
# caller is calling routine - if provided will write debug statements
# filehandle is opened filehandle
# filename is name of the opened file
#
# TODO: Need to support netwide locking for automated CQ transaction 
# posting and potentially for UNIX users using multiple hosts but
# not centralized caches.
######################################################################

my $FLOCK_SHARED    = 1;    # flock codes to lock files
my $FLOCK_EXCLUSIVE = 2;
my $FLOCK_NONBLOCK  = 4;
my $FLOCK_UNLOCK    = 8;

sub FileLock {
    my ($self, $operation, $caller, $fileHandle, $filename) = @_;

    if ($operation ne $IO_FLOCK_UNLOCK) {
        my $mode = ($operation eq $IO_FLOCK_READ) ? $FLOCK_SHARED : 
                   ($operation eq $IO_FLOCK_WRITE) ? $FLOCK_EXCLUSIVE :
                   undef;

        if (! $mode) {
            $self->IntMessage("FATAL", 
                    ["IO_E_FAILED_FILE_LOCK", $operation, $filename]);
            return;
        }

        if (! -e $filename) {
            $self->IntDebugPrint("VERY_DETAILED", 
                "$caller ($filename) $operation lock requested on " . 
                "non-existent file");
            return;
        }

        if (flock($fileHandle, $mode|$FLOCK_NONBLOCK)) {
            # TODO: DebugPrint might ALSO hit locks, etc. 
            if ($caller && $filename) { 
                my $time = localtime(time);
                $self->IntDebugPrint("VERY_DETAILED", 
                    "$caller ($filename) $operation Lock obtained " . 
                    "first try ($time)");
            }
        } else {
            if ($caller && $filename) {
                my $time = localtime(time);
                $self->IntDebugPrint("VERY_DETAILED", 
                    "$caller ($filename) Waiting $operation Lock ($time)");
            }
            if (! flock($fileHandle, $mode)) {
                $self->IntMessage("FATAL", 
                    ["IO_E_FAILED_FILE_LOCK", $mode, $filename]);
                return;
            }
            if ($caller && $filename) {
                my $time = localtime(time);
                $self->IntDebugPrint("VERY_DETAILED", 
                    "$caller ($filename) $operation Lock obtained after " . 
                    "wait ($time)");
            }
        }
        # Backdoor for testing to hold locks arbitrarily long
        if ($ENV{"CQCC_FLOCK_SLEEP"}) {
            sleep($ENV{"CQCC_FLOCK_SLEEP"});
        }
    } else {
        close($fileHandle);
        if ($caller && $filename) {
            my $time = localtime(time);
            $self->IntDebugPrint("VERY_DETAILED", 
                   "$caller ($filename) Lock released ($time)");
        }    
    }
    return;
}

######################################################################
#::function IOFileOperation

=over 6

=item C<IOFileOperation(operation, type, caller, filename, newName)>

This routine provides a platform independent wrapper for basic
file operations. If the operation fails, the error will be
reported at the type level as coming from caller

Supported Operations
* $IO_FILE_COPY - copy filename to newName
* $IO_FILE_RENAME -  rename filename to newName

=back 

=cut

# 
######################################################################

sub IOFileOperation {
    return &IOstd->FileOperation(@_);
}

######################################################################
#::method FileOperation
######################################################################

sub FileOperation {
    my ($self, $operation, $type, $caller, $filename, $newName) = @_;
    my $cmd;
    if ($operation eq $IO_FILE_COPY) {
        if ($self->{WINDOWS}) {
            $cmd = "copy \"$filename\" \"$newName\"";
        } else {
            $cmd = "cp $filename $newName";
        }
    } elsif ($operation eq $IO_FILE_RENAME) {
        if (-e $newName) {
            unlink $newName;
        }
        if ($self->{WINDOWS}) {
            my $base = basename($newName);
            if (&GetParm("CM_WIN_UNIX") eq "FALSE") {
                $cmd = "rename \"$filename\" \"$base\"";
            } else {
                rename $filename,$newName;
                if (! -e $newName) {
                    $self->IntMessage($type,
                        [ "IO_E_FAILED_FILE_OP", $operation, $filename, 
                          $newName, $caller ]);
                    return 0;
                }
                return 1;
            }
        } else {
            my $dirOld = dirname($filename);
            my $baseOld = basename($filename);
            my $dirNew = dirname($newName);
            my $baseNew = basename($newName);
            if ($dirOld eq $dirNew) {
                $cmd = "cd $dirOld; mv $baseOld $baseNew";
            } else {
                $cmd = "mv $filename $newName";
            }
        }
    } else {
        $self->IntMessage($type,
            [ "IO_E_INVALID_FILE_OP", $operation, $filename, $newName, $caller ]);
    }
    if (system($cmd)) {
        # TODO: Improve error reporting
        $self->IntMessage($type,
            [ "IO_E_FAILED_FILE_OP", $operation, $filename, $newName, $caller ]);
        return 0;
    }
    return 1;
}

######################################################################
#::function IOFileWrite

=over 6

=item C<IOFileWrite(filename, item(s))>

This accessor function opens filename, writes the item list to it
and closes the file. Each value is written out as a single line in
the file. While not a broad purpose file writer, it is intended to
work in tandem with FileRead which reads one item to a line.

=back 

=cut

######################################################################

sub IOFileWrite {
    return &IOstd->FileWrite(@_);
}

######################################################################
#::method FileWrite
######################################################################

sub FileWrite {
    my ($self, $filename, @args) = @_;
    return $self->FileOutput($filename, $IO_FILE_WRITE, @args);
}

######################################################################
#::method FileOutput
######################################################################

sub FileOutput {
    my ($self, $filename, $mode, @args) = @_;
    if (($mode ne $IO_FILE_WRITE) && ($mode ne $IO_FILE_APPEND)) {
        # TODO: Improve error reporting
        $self->IntMessage("FATAL", 
            ["IO_E_INVALID_FILE_OUTPUT", $mode, $filename]);
        return;
    }
    # Strip off enclosing quotes
    if (substr($filename,0,1) eq '"') {
        $filename = substr($filename,1,length($filename)-2);
    } 

    open (OUTFILE, "$mode$filename");
    $self->FileLock($IO_FLOCK_WRITE, "IOBase::FileOutput",\*OUTFILE,$filename);

    for (@args) {
        print OUTFILE $_, "\n";
    }      
    $self->FileLock($IO_FLOCK_UNLOCK, "IOBase::FileOutput",\*OUTFILE,$filename);

    # Mark known temp files for deletion later
    if (defined $TempFiles{$filename}) {
        $TempFiles{$filename} = 2;
    } 
}

######################################################################
#::function IOHelpDisplay

=over 6

=item C<IOHelpDisplay(filename, topic)>

This accessor function will launch a help browser for a specific
HTML help file. 

=back 

=cut

######################################################################

sub IOHelpDisplay {
    return &IOstd->HelpDisplay(@_);
}

######################################################################
#::method HelpDisplay
######################################################################

sub HelpDisplay {
    my ($self, $filename, @args) = @_;

    # Strip off enclosing quotes
    if (substr($filename,0,1) eq '"') {
        $filename = substr($filename,1,length($filename)-2);
    } 
    # In Windows, rely on type manager to associate html suffix
    # with preferred web browser
    my $launcher = $self->{HELPAPP}; 
    my $cmd = ($launcher ne "" ? "$launcher " : "") . $filename;
    $self->IntDebugPrint("BASIC", 
        "Launching help display using \n  $cmd\n");
    system($cmd);
}

######################################################################
#::function IOInterruptHandler

=over 6

=item C<IOInterruptHandler()>

When enabled by the CQCC_CTRL_C configuration parameter, this function
catches Contrl-C interrupts and calls IOExitHandler to force an 
operation cancel. 

=back 

=cut

######################################################################

sub IOInterruptHandler {
    # Don't accept interrupt before IOBase initialization is completed
    if (&IOInitialized()) { 
        &IODebugPrint("BASIC", "Control-C invoked, " . 
            "calling IOExitHandler\n");
        &IOExitHandler(2);
    }
    return;
}

######################################################################
#::function IOListSelect

=over 6

=item C<IOListSelect(multiChoice, titleMsgTag, choiceMsgTagList...)>

This accessor function presents a menu with a title described by
"titleMsgTag" and one or more options to select from described by the
remaining arguments. If multiChoice is 1 then multiple choices are
allowed, otherwise only one choice is provided.

In the text implementation, the choices are numbered on the left
margin and the user is expected to type one or more numbers separated
by any combination of spaces or commas.

The selected items will be returned as a list to the caller, 
using the original choiceMsgTag items in their entirety for the caller
to match to expectations. 

The return value consists of a list containing the a flag to abort
(1 for abort, 0 for continue) and the answer(s).

=back 

=cut

######################################################################

sub IOListSelect {
    return &IOstd->ListSelect(@_);
}

######################################################################
#::method ListSelect
######################################################################

sub ListSelect {
    my ($self, $multiChoice, $titleMsgTag, @choices) = @_;

    my (@answers,$i,$done,$choice, $msg, $menuLine,$maxItems);
    
    $self->_LogEntry("ListSelect", $titleMsgTag);

    $msg = $self->{MSGCAT}->Message($titleMsgTag);

    &IOPrint("\n$msg\n");
    $i=1;
    for $choice (@choices) {
        $menuLine = $self->{MSGCAT}->Message($choice);
        &IOPrint("($i) $menuLine\n");
        $i++;
    }
    $maxItems = $i-1;

    $done = 0;
    while (! $done) { 
        $done = 1;

        my $answer = $self->_ReadSTDIN( "DURING",
           ["IO_P_LISTSELECT", ($multiChoice ? "(s)" : "")]);
      
        $answer =~ s/,/ /g;
        my @answerList = &IOParseListString($answer);
        @answers = undef;
        my $count = 0;
        my $a;
        for $a (@answerList) {
            if (($a > 0) && ($a <= $maxItems)) {
                $answers[$count++] = $choices[$a - 1];
            } else {
                &IOMessage("WARN", ["IO_E_ANSWER_RANGE", $a]);
                $done = 0;
            }
        }
        if (! $multiChoice && $#answerList > 0) {
            &IOMessage("WARN", ["IO_E_SINGLE_ANSWER"]);
            $done = 0;
        }
    }

    return (0, @answers);
}

######################################################################
#::function IOListTreeSelect

=over 6

=item C<IOListTreeSelect(multiChoice,separator,titleMsgTag,choiceMsgTagList...)>

Variant on IOListSelect which presents the list as a folding tree
structure based on the choice list, assuming that each choice is a
pathname separated by the separator character. 

If the IOBase class does not provide support for the IOListTreeSelect
method, it will default to using the standard IOListSelect method
instead and present the information as a normal list.

For more details see the IOListSelect function documentation. 

=back 

=cut

######################################################################

sub IOListTreeSelect {
    return &IOstd()->ListTreeSelect(@_);
}

sub ListTreeSelect {
    my ($self, $multiChoice, $sep, $titleMsgTag, @choiceList)=@_;
    # Drop separator arg before pass through. 
    return $self->ListSelect($multiChoice, $titleMsgTag, @choiceList);
}

######################################################################
#::function IOMessage

=over 6

=item C<IOMessage(type, abort, [msgtag, args])>

This accessor function will display its message to the user, after
translating the msgtag and its arguments into the corresponding
message catalog entry. Display is handled by calling the Prompt()
method.

The type value may be "INFO", "WARN", "ERROR", or "FATAL" and is
intended to qualify how the message is to be shown and whether or
not the message affects the overall trigger status. If "FATAL" the
status will be set to failed and the trigger will try to cleanly
unwind and report failure back to the original caller, usually
ClearCase.

The abort flag is also used to signal general failure for terminal
conditions and will be combined with the type information shortly. 

=back 

=cut

######################################################################

sub IOMessage {
    my ($type, $msgTag) = @_;
    my $result = &IOstd->Message(@_);
    if ($type eq "ERROR" || $type eq "FATAL") {
        &IOstd->{ERR_COUNT} = &IOstd->{ERR_COUNT} + 1;
    }
    return $result;
}

######################################################################
#::method Message
######################################################################

sub Message {
    my ($self, $type, $msgTag) = @_;
    
    my ($package, $filename, $line) = caller 1;
    my $proceed = 1;
    my $result = 1;

    if (! defined $msgTag) {
        # Avoid using subclasses
        $self->IntMessage("ERROR", ["IO_E_MSGTAG_UNDEF"]);
        IOTraceBack("Message", 10);
    } else {
        my $msgType = $self->{MSGCAT}->Message([$type]);
        my $msg = $self->{MSGCAT}->Message($msgTag) . "\n";
        $result = $self->_LogEntry("Message", $msgTag, "1");
    
        $self->Print("$msgType: $msg\n");
        if ($type eq "ERROR" || $type eq "FATAL") {
            # This is for special internal test purposes only.
            # It is not recommented for general use.

            if ( defined($ENV{"CQCC_TEST_USER"})) {
                $self->PrintStream("ERROR", "Caller: $filename Line: $line\n");
            } else {
                $self->Print("Caller: $filename Line: $line\n");
            }
        }

        # Record message to error log if enabled
        $self->MessageLog($type,$msgTag,$filename,$line);

        # if ($type eq "ERROR") {
        # TODO: Modify to new YesNo dialog if revived
        #    my $answer = $self->_ReadSTDIN("DURING", 
        #        ["IO_P_MESSAGE" ]);
        #    if (uc(substr($answer,0,1) eq "N")) {
        #        $proceed = 0;
        #    }
        # }
    }    
    if ($type eq "FATAL" || $result eq "0") {
        $self->ExitHandler(2);
    } elsif ($proceed == 0) {
        $self->ExitHandler(1);
    }

    return;
}

######################################################################
#::method IntMessage - Internal Message routine
# Conditionally call IOMessage function if fully initialized
# else call IOBase Message method directly for initialization errors
######################################################################

sub IntMessage {
    my ($self,@args) = @_;
    if (&IOInitialized()) { 
        &IOMessage(@args);
    } else {
        # Message Queuing - If IO object not yet initialized in CCRC 
        # delay presenting messages until initialization is complete
        # to ensure the user sees the messages. 
        if ($ENV{"ATRIA_WEB_GUI"}) {
            my @delayedMessage = @args;
            my $msg_queue = $self->{MSG_QUEUE};
            my @msgQueue = (defined $msg_queue) ? @{$msg_queue} : ();
            push @msgQueue, \@delayedMessage;
            $self->{MSG_QUEUE} = \@msgQueue;
        } else {
            &CQCC::IOBase::Message($self, @args);
        }
    }    
}

######################################################################
#::method MessageLog
######################################################################

sub MessageLog {
    my ($self, $type, $msgTag, $filename, $line) = @_;
    my ($level,$errorFlags) = split(":",&GetParm("CQCC_LOG_ERRORS"));
    $errorFlags = "*"; # Force all parameters to always be shown
    my $errorLogLevel = $messageLevels{$level};
    if ($self->{ERR_FILE} && $errorLogLevel ne 0 && 
            $messageLevels{$type} >= $errorLogLevel) {
        # TODO: For I18N, specify local code page for file output
        my $msg = $self->{MSGCAT}->Message($msgTag) . "\n";
        open (MSGFILE, ">>$self->{ERR_FILE}");
        # Get exclusive write lock, no debug output
        $self->FileLock($IO_FLOCK_WRITE, undef, \*MSGFILE, $self->{ERR_FILE});
        my $bar = "::##################################################";
        my $time = localtime(time);
        print MSGFILE "$bar\n$time\n";
        print MSGFILE "$type: $msg\n--------\n";

        if ($errorFlags) {
            my @printParms;
            if ($errorFlags eq "*") {
                @printParms = &GetParmNameList();
            } else {
                @printParms = split(",",$errorFlags);
            }
            for (sort @printParms) {
                print MSGFILE &ParmString($_) . "\n";
            }
        }
        if ($type eq "ERROR" || $type eq "FATAL") {
            print "Caller: $filename Line: $line\n";
        }
        $self->FileLock($IO_FLOCK_UNLOCK, undef, \*MSGFILE, $self->{ERR_FILE});
    }
}

######################################################################
#::function IOLockOperation

=over 6

=item C<result = IOLockOperation(name,lock,optionsList)>

This method places a cross-thread lock on the named operation 
by opening a file called "name.LOCK" if it there isn't one
or waiting until the file doesn't exist anymore before proceeding.

The name.LOCK file has a unique message written to it to identify
the process that created the lock. This can be used to confirm that 
there has been no attempt to break the lock since the lock was
created and to confirm that this process has lock ownership when 
it attempts to delete the file.

Optional arguments allow the lock to be automatically broken 
when it has been left around beyond a specified time by another
process. 

Arguments: 
    name must be a valid pathname

    lock is $IO_LOCKOP_NET (net safe lock)
         or $IO_LOCKOP_VERIFY (confirm lock unbroken)
         or $IO_LOCKOP_UNLOCK(unlock)

    optionsList is a list of one or more modifier pairs
         "AUTOBREAK_MINS" => automatically break existing lock
                             if more than this many minutes old
                             Return error if unable to break lock
         "WAIT_WARN_MINS"  => wait indefintely until file disappears 
                             or we can break the lock. Warn user after
                             this many minutes. 

When a lock is no longer needed, the unlock operation should be 
called to release the lock. Locks will automatically be released
when the process ends and the lock file is automatically closed. 

Returns 0 if operation failed, else 1

=back 

=cut

######################################################################

sub IOLockOperation {
    return &IOstd->LockOperation(@_);
}

######################################################################
#::method LockOperation
######################################################################

sub LockOperation {
    my ($self,$name,$lockOp,%optionsList) = @_;

    # If unlock/no-name, release all current locks
    if (! $name) {
        if ($lockOp eq $IO_LOCKOP_UNLOCK) {
            for (keys %opLocks) {
                $self->LockOperation($_,$lockOp);
            }
        }
        return 1;
    }

    my $fileName = $name . ".LOCK";
    if ($lockOp eq $IO_LOCKOP_NET) {
        # Get lock by waiting for name.LOCK file to not exist then
        # create it with unique contents to signal ownership. 

        my $autobreak = (defined $optionsList{"AUTOBREAK_MINS"}) 
                        ? $optionsList{"AUTOBREAK_MINS"} : 0;
        my $breakAge = $autobreak * 60;
        my $autobreakWait = (defined $optionsList{"WAIT_WARN_MINS"}) 
                        ? $optionsList{"WAIT_WARN_MINS"} : 0;
        my $fileModifiedDate;

        while (-e $fileName) { 
            if (! defined $fileModifiedDate) {
                my @fileInfo = stat($fileName);
                $fileModifiedDate = $fileInfo[9];
            }
            # Automatically delete stale locks over specified age
            # if AUTOBREAK option is enabled

            if (-e $fileName && $autobreak) {
                if ((time() - $fileModifiedDate) > $breakAge) {
                    &IODebugPrint("DETAILED", "Breaking lock $fileName\n");
                    unlink $fileName;
                }
                # If file not stale enough or delete failed, error
                if (-e $fileName && ! $autobreakWait) {
                    return 0;
                }
            }

            if (-e $fileName) {
               $self->IntDebugPrint("DETAILED", 
                   "LockOperation waiting for $fileName to be released");
               my $maxWait = $autobreakWait * 60; 
               my $i;
               for ($i = 0; $i < $maxWait; $i++) {
                   sleep(1);
                   if (! -e $fileName) {
                       $self->IntDebugPrint("DETAILED", 
                           "LockOperation $fileName became available");
                       last;
                   } elsif ($autobreak && ((time() - $fileModifiedDate) > $breakAge)) {
                        &IODebugPrint("DETAILED", "Breaking lock $fileName\n");
                        unlink $fileName;
                   } elsif ($i >= ($maxWait-1)) {
                       $self->IntMessage("INFO", ["IO_P_AUTOBREAK_WAIT"]);
                       last;
                   }
               }
            }
        }
        my $lockMessage = "IOLockOperation: User(" . $self->{LOCALUSER} . 
                  ") Host(" . $self->{LOCALHOST} . 
                  ") Trigger (" . &GetParm("TriggerInterface") . 
                  ") Process ($$)";
        $self->FileWrite($fileName,$lockMessage);
        $opLocks{$name} = $lockMessage;

        # If lock isn't verified then release it and try again. 
        # Should only be encountered if two threads try to grab lock
        # simultaneously. 
        if (! &IOLockOperation($name, $IO_LOCKOP_VERIFY)) {
            # If lock not verified, release it and try again
            &IOLockOperation($name, $IO_LOCKOP_UNLOCK);
            return &IOLockOperation($name,$lockOp,%optionsList);
        }

        return 1;
    } else {
        # Lookup name and message 
        # Verify message against file contents
        # If unlocking delete entry and delete lock file if verified
        # Otherwise depend on autobreak to delete file later on. 
        if ($opLocks{$name}) {
             my $lockMessage = $opLocks{$name};
             my ($fileMessage) = &IOFileRead($fileName);
             my $sameMessage = ($fileMessage eq $lockMessage) ? 1 : 0;
             if (! $sameMessage) {
                 &IODebugPrint("BASIC", "IOLockOperation: Lock broken for $fileName\n" .
                     "Expected lock: $lockMessage\n" . 
                     "Found lock: $fileMessage");
             }
             if ($lockOp eq $IO_LOCKOP_VERIFY) {
                 return $sameMessage;
             } elsif ($lockOp eq $IO_LOCKOP_UNLOCK) {
                 delete $opLocks{$name};
                 if ($sameMessage) {
                     unlink $fileName;
                 }
             }
        }
        return 1;
    }
}

######################################################################
#::function IOParseListString

=over 6

=item C<IOParseListString()>

This utility function takes a string with embedded commas and white
space and parses it into a list.

=back

=cut

######################################################################

sub IOParseListString {
    my $str = shift;

    $str =~ s/,/ /g;
    return split(" ",$str);
}

######################################################################
#::function IOPathname

=over 6 

=item C<IOPathname($type, $filename)>

This accessor function is used to convert platform independent
filenames into platform specific pathname strings in specific
directories based on the $type value. The $type may also cause
a suffix to be added to further qualify the name by adding
a process id or series id to make the name unique.

Supported type fields:
* ATRIA - filename in Clearcase home directory
* PARM - filename in parameter directory used to cache info
* PARM_USER - filename in original (non-centralized) parameter home dir
* PARM_SERIES - filename in parameter directory with series id suffix
* TEMP - filename in temporary directory with process id included

Based on the operating system, the type directories are set by the
Initialize() function and the remaining pathname is modified to adjust
the slash (/ or \) conventions appropriately.

Note that TEMP files will be tracked by the integration and 
automatically deleted on exit. 

=back 

=cut

######################################################################

sub IOPathname {
    return &IOstd->Pathname(@_);
}

######################################################################
#::method Pathname
######################################################################

sub Pathname{
    my ($self, $type, $path) = @_;
    my $prefix = defined $self->{$type} ? $self->{$type} : 
                                          "<unknown_prefix>";
    my $pathname = $prefix . "/" . $path;

    if ($self->{WINDOWS}) { 
        $pathname =~ s#\/#\\#g;
        if ($type eq "ATRIA" && $pathname =~ /\s/) {
           $pathname = '"' . $pathname . '"';
        }
    } else {
        $pathname =~ s#\\#/#g;
    } 

    # Remember temp files - if used later, cleanup on exit
    if ($type eq "TEMP") { 
        # Add process id qualifier for unix (more likely to be multiuser)
        if (! $self->{WINDOWS}) {
            $pathname .= ".$$";
        }
        else {
            my $ext = "";
            if ($pathname =~ s/(\.\w+)$//) {
                $ext = $1;
            } 
            $pathname .= ".$$" .$1;    
        }
        if (! defined $TempFiles{$pathname}) {
            $TempFiles{$pathname} = 1; 
        } 
    } elsif ($type eq "PARM_SERIES" && $self->{SERIES_KEY}) { 
        my $base = basename($pathname);
        # If file name is form of "something.extension"
        # insert the series id before the extension
        if ($base =~ /(\S+)\.(\S+)$/) {
            my $dir = dirname($pathname);
            $pathname = $dir . $sep .
                        "$1." . $self->{SERIES_KEY} . ".$2";
        } else {
            $pathname .= "." . $self->{SERIES_KEY};
        }
    }
    return $pathname;
}

######################################################################
#::function IOPrint

=over 6 

=item C<IOPrint(print-args)>

This accessor function handles informational print operations,
allowing prints to be redirected to a log as well as stdout.

=back 

=cut

######################################################################

sub IOPrint {
    return &IOstd->Print(@_);
}

######################################################################
#::method Print
######################################################################

sub Print{
    my ($self, @args) = @_;
    return PrintStream($self, "INFO", @args);
}   

######################################################################
#::function IOPrintStream

=over 6 

=item C<IOPrintStream(stream, print-args)>

This accessor function handles informational print operations,
allowing prints to be redirected to a log as well as a specified stream
as indicated by the first argument.

=back 

=cut

######################################################################

sub IOPrintStream {
    return &IOstd->PrintStream(@_);
}

######################################################################
#:: method PrintStream
######################################################################

sub PrintStream{
    my ($self, $type, @args) = @_;

    $self->PrintLog($IO_FILE_APPEND,@args);
    if ($type eq "ERROR") {
        print STDERR @args ;
    } else {
    print @args;
}
}

######################################################################
#::method PrintLog
######################################################################

sub PrintLog {
    my ($self, $mode, @args) = @_;
    if (&GetParm("CQCC_LOG_OUTPUT") ne "OFF" && $self->{LOG_FILE}) { 
        open (LOGFILE, $mode . $self->{LOG_FILE});
        $self->FileLock($IO_FLOCK_WRITE, undef, \*LOGFILE, $self->{LOG_FILE});
        print LOGFILE @args;
        $self->FileLock($IO_FLOCK_UNLOCK, undef, \*LOGFILE, $self->{LOG_FILE});
    }
}

######################################################################
#::function IOPrompt

=over 6

=item C<IOPrompt(promptMessage)>

This accessor function will display its message to the user, after
translating the msgtag and its arguments into the corresponding
message catalog entry.

The return value consists of a list containing a flag to abort
(1 for abort, 0 for continue) and the answer.

($abort, $answer) = IOPrompt(promptMessage);

=back

=cut

######################################################################

sub IOPrompt {
    return &IOstd->Prompt(@_);
}

######################################################################
#::function IOPromptLogin

=over 6

=item C<IOPromptLogin(textPromptMessage, dialogPromptMessage)>

This accessor function displays a login dialog requesting the user
name and password. 

User interfaces that have to prompt for both values in one entry
window separated by a space such as the command line (IOBase) or
ClearPrompt (IOClearPrompt) use the textPromptMessage as the input
prompt.

User interfaces that can present the user name and password entries
separately in a dialog use the dialogPromptMessage as the prompt.

The return value consists of a list containing a flag to abort
(1 for abort, 0 for continue) and the name and password.

($abort, $username, $password) = IOPromptLogin(textPrompt,dialogPrompt)

=back

=cut

######################################################################

sub IOPromptLogin {
    my ($textPromptMessage, $dialogPromptMessage) = @_;
    return &IOstd->Prompt($textPromptMessage,
        "IOPROMPT_MODE" => "IOPROMPT_LOGIN",
        "IOPROMPT_LOGIN_MSG" => $dialogPromptMessage);
}

######################################################################
#::function IOPromptYesNo

=over 6

=item C<IOPromptYesNo(promptMessage)>

This accessor function displays a login dialog requesting the answer
to a yes or no question. 

User interfaces that have to prompt the user for text input  
such as the command line (IOBase) will automatically add the valid 
answers, i.e. "(y/n)" based on the local message catalog. 

User interfaces that can present yes and no buttons will use the 
original promptMessage without modification. 

The user's input is sent to the current message catalog to be 
interpreted as a yes or no. 

The return value consists of a list containing a flag to abort
(1 for abort, 0 for continue) and the answer (1 for yes, 0 for no). 

($abort, $answer) = IOPromptYesNo(promptMessage)

=back

=cut

######################################################################

sub IOPromptYesNo {
    my ($promptMessage) = @_;
    return &IOstd->Prompt($promptMessage,
        "IOPROMPT_MODE" => "IOPROMPT_YES_NO");
}

######################################################################
#::method Prompt
 
=over 6

=item C<Prompt(promptMsg, optionsList)>

This internal function displays its message to the user, after
translating the msgtag and its arguments into the corresponding
message catalog entry. In addition to the required message, it 
accepts optional qualifiers in the form of "option=>value" pairs to
support multiple public API calls like IOPromptLogin.

Available options include the following: 

   "IOPROMPT_MODE" =>     # Prompt display mode, one of the following

       "IOPROMPT_TEXT"   # Default, display message, expect user input 

  or   "IOPROMPT_LOGIN"  # Login prompt, expect user to enter username and (masked) password

  or   "IOPROMPT_YES_NO" # Present user choice of yes or no input

   "IOPROMPT_LOGIN_MSG"=>loginPromptMessage  # Alternate login prompt

The return value consists of a list containing the abort status
(1 for abort, 0 for continue) and the answer(s). 

"IOPROMPT_LOGIN" returns two strings, username and password.
"IOPROMPT_YES_NO" returns a single integer, 1 for yes, 0 for no.

($abort, $answer(s)) = $obj->Prompt(msg, option1=>value1, ... );

=back

=cut

######################################################################

sub Prompt {
    my ($self, $msgTag, %options) = @_;
    my $mode = (defined $options{"IOPROMPT_MODE"}) ? 
               $options{"IOPROMPT_MODE"} : "IOPROMPT_TEXT";

    my $msg = $self->{MSGCAT}->Message($msgTag);

    # LOGIN: Just use default prompt, return name, password
    # TODO: When input masking becomes available, implement two step
    # login prompt, masking password input. 

    my $answer;
    my $abort = 0;

    $self->_LogEntry("Prompt", $msgTag);

    # If mode is yesNo, append "($yes / $no)" to end. 
    if ($mode eq "IOPROMPT_YES_NO") {
        $msg .= " " . $self->{MSGCAT}->Message([ "IO_P_YES_NO" ]);
    }
    &IOPrint("$msg\n");
    $answer = $self->_ReadSTDIN("DURING", [ "IO_P_PROMPT" ] );
    if ($mode eq "IOPROMPT_YES_NO") {
        my $yesNo = $self->{MSGCAT}->YesNo($answer);
        if ($yesNo eq 0) {
            $abort = 1;
        } else {
            $answer = ($yesNo eq 1) ? 1 : 0; 
        }
    } elsif ($mode eq "IOPROMPT_LOGIN") {
        my ($username,$password) = split(/[\s,;:]/, $answer, 2);
        return ($abort, $username, $password);
    } 
    return ($abort, $answer);
}

######################################################################
#::function IOQuotePathname

=over 6 

=item C<IOQuotePathname(pathname)>

This utility function will add quotes around pathnames only if needed
for handling embedded whitespace or caret characters. Otherwise the 
original pathname is returned. 

=back 

=cut

######################################################################

sub IOQuotePathname {
    my ($path) = @_;
    my $char = '$';
    my $result = index($path, $char);
    if ($result > 0)  {
        return "\'" . $path . "\'" ;
    } else {
        return "\"" . $path . "\"" ;
      }
}

######################################################################
#::function IOResourceManager

=over 6

=item C<IOResourceManager(resourceName,operation,maxAgeMins)>

Provide generic reuseable resource file management.

resourceName should be base name of the resource file to be 
managed. 

Operations: 
* 1 - Acquire available resource using base name
* 0 - Release resource file using base name
      If no basename is provided, release ALL resources

Returns fully qualified parameter filename based on resourceName
with series suffix added. Name is stored for later look up 
based on the resourceName. 

If maxAgeMins is non-zero, obsolete resource files older than 
that age are deleted if they are seen before a valid resource is 
found. 

=back 

=cut

######################################################################

sub IOResourceManager {
    return &IOstd->ResourceManager(@_);
}

######################################################################
#::method ResourceManager
######################################################################

sub ResourceManager {
    my ($self,$resourceName,$operation,$maxAgeMins) = @_;
    my $maxAgeSecs = $maxAgeMins * 60;

    # Return if IO object not fully initialized
    if (! &IOInitialized()) { 
        return;
    }

    &IODebugPrint("DETAILED", "ResourceMgr " .
                      ($operation ? "acquiring " : "releasing ") . 
                      ($resourceName ? $resourceName : "ALL"));
    if ($operation eq 1) {
        # If resource already acquired, return that name again
        # Does NOT support obtaining multiple resources with same name
        my $resName = $managedResources{$resourceName};
        if ($resName) {
            return $resName;
        }

        # Have all resource managers lock a common (unused) file
        # to avoid multiple threads competing for same resources

        my $resMgrFile = &IOPathname("PARM",".cqcc_resource_mgr");
        my ($file,$found); 

        my $acquiredFile = &IOPathname("PARM_SERIES",$resourceName);

        # If unable to obtain or break the lock on the pool
        # just create a new resource and add it back into the pool later

        my $lock = &IOLockOperation($resMgrFile,$IO_LOCKOP_NET,
                                    "AUTOBREAK_MINS" => &GetParm("CM_LOCK_BREAK_MINS"));

        # If can't get lock on first attempt, try once more after waiting 1 second
        # to minimize impact of collisions. 
        if (! $lock) {
            sleep(1);
            $lock = &IOLockOperation($resMgrFile,$IO_LOCKOP_NET,
                                     "AUTOBREAK_MINS" => &GetParm("CM_LOCK_BREAK_MINS"));
        }

        if ($lock) {
            # Acquire: Search for base.*.avail files
            my $resDirectory = &IOPathname("PARM","."); 
            opendir RESDIR, $resDirectory;
            my @allFiles = readdir RESDIR;
            closedir RESDIR;

            # Prefer to find own connection file
            
            if (-e $acquiredFile . ".avail") {
                $found = $acquiredFile . ".avail";
                # If preferred file is stale, skip it and let it be deleted 
                # in main search loop
                if ($maxAgeSecs > 0) {
                    my @fileInfo = stat($found);
                    if ((time() - $fileInfo[9]) > $maxAgeSecs) {
                        $found = undef;
                    }
                }
            }

            if (!defined $found) {
                for $file (@allFiles) { 
                    if ($file =~ /$resourceName\.(\S*)\.avail/) {
                        # Flush all stale connection files when checking resources
                        # If maxAgeSecs provided, delete obsolete avail resource
                        if ($maxAgeSecs > 0) {
                            my $fullPath = $resDirectory . "/" . $file;
                            my @fileInfo = stat($fullPath);
                            if ((time() - $fileInfo[9]) > $maxAgeSecs) {
                                &IODebugPrint("DETAILED",
                                    "IOResourceManager: Deleting $file " . 
                                    " age exceeds $maxAgeMins minutes");
                                unlink($file);
                                next;
                            }
                        }
                        # Grab first available file for reuse
                        if (! defined $found) {
                            $found = $file;
                            $found = &IOPathname("PARM",$found); 
                        }
                    }
                }
            }        
        }

        # If found, then rename base.*.avail to base.SERIES_ID
        # If not found or pool locked, return name of new empty file
        if ($found) {
            &IODebugPrint("DETAILED",
                "IOResourceManager: Acquiring $file for reuse");
            &IODebugPrint("DETAILED", "ResourceMgr: Acquiring $found");
            &IOFileOperation($IO_FILE_RENAME, 
                             "ERROR", "IOResourceManager",   
                             $found,$acquiredFile);
        }

        # Add name to resource hash
        $managedResources{$resourceName} = $acquiredFile;
        &IOLockOperation($resMgrFile,$IO_LOCKOP_UNLOCK);
        return $acquiredFile;

    } else {
        # Lock not required since nobody else should have access 
        # to managed resource files and rename operation is atomic

        if (! $resourceName) { 
            # Release any managed resources 
            for (sort keys %managedResources) {
                &IOResourceManager($_, 0, 1);
            }
        } else {
            # Release: Release specific named resource
            # Only rename if file exists, otherwise doesn't matter. 
            # If acquired resource deleted, it can't be made available. 

            my $resName = $managedResources{$resourceName};
            if ($resName && -e $resName) {
                my $newName = $resName . ".avail";
                # Clear resource before attempting to rename so 
                # if there are errors, we don't keep retrying to release
                delete $managedResources{$resourceName};
                &IOFileOperation($IO_FILE_RENAME, 
                                 "ERROR", "IOResourceManager",   
                                 $resName,$newName);
                &IODebugPrint("DETAILED",
                    "IOResourceManager: Releasing $resName to reuse " . 
                    " as $newName");
            }
        }
        return "";
    }
}

######################################################################
#::function IOGetCleanupHandler

=over 6

=item C<(object,method) = IOGetCleanupHandler()>

This function returns the cleanup handler information set
by IOSetCleanupHandler.

=back

=cut

######################################################################

sub IOGetCleanupHandler {
    return ($_ExitObject, $_ExitMethod);
}

######################################################################
#::function IOSetCleanupHandler

=over 6

=item C<IOSetCleanupHandler(object, method)>

This function stores (or clears) the command string to be evaluated
during exits processed by IOExitHandler(). The object is a pointer
to the object whose method will be called. The method string should 
be either "Method()" or "Method(args)". On exit the command 
"object->method" will be invoked as a cleanup handler. 

=back

=cut

######################################################################

sub IOSetCleanupHandler {
    $_ExitObject = shift;
    $_ExitMethod = shift;
}

######################################################################
#::function IOTraceBack

=over 6

=item C<IOTraceBack()>

This is a generic method to print out the calling stack back to N
levels. It should be moved to a more generic base class for
all objects when time permits. 

=back

=cut

######################################################################

sub IOTraceBack{
    my ($label, $levels) = @_;
    if (! defined $levels) {
        $levels = 4;
    }     
    my ($i, $funName, $pkg, $lineNo);
    my $msg = "";
    &IOPrint("IOTraceBack: $label\n");
    for ($i = 0; $i < $levels; $i++) { 
        my ($package, $filename, $line, $func) = caller $i;
        if ($i > 0) {
            &IOPrint("   ($i) $pkg::$func() line: $lineNo\n");
            $msg .= "   ($i) $pkg::$func() line: $lineNo\n";
        }
        $pkg = $package;  # use pkg and linenum with next func
        $lineNo = $line;  #    to report who did calling at that level
    }
    my ($package, $filename, $line) = caller 1;
    &IOstd()->MessageLog("ERROR", "IOTraceback:\n" . $msg, $filename,$line);
}

######################################################################
#::method new

=over 6

=item C<new CQCC::IOBase()>

This method creates the basic object instance and presets its
attribute values. Once created, the caller needs to invoke 
the Initialize() method for more extensive initialization. 

=back

=cut

######################################################################

sub new { 
    my $proto = shift;
    my $class = ref($proto) || $proto;  # Object or class name

    # Prompt strings moved to MsgCat_English.pm

    # Attributes
    my $self = {
        CLASSNAME  => "CQCC::IOBase",
        FAILED => 0,          # Trigger status, false = exit-we-failed
        MSGCAT => undef,      # Message catalog object
        CLEARPROMPT => undef, # ClearPrompt executable for dialog box
        DEVNULL => undef,     # Where to send undesired output
        WINDOWS => undef,     # Running on Windows(1) or UNIX(0) 
        HELPAPP => "",        # Executable needed to launch HTML help pages
        JOURNAL_MODE => "OFF",# Controls use of journaling
        LOG_FILE => undef,    # Log file for output when enabled
        ERR_FILE => undef,    # Log file for error output
        ERR_COUNT => 0,       # Number of errors reported
        INITIALIZED => undef, # Flag when initialization is completed
        MSG_QUEUE => undef,   # List of delayed delivery messages 

        # L10N: Remember desired codepage, CORE by default

        IO_CODEPAGE => "CORE_CODEPAGE",   # Assume desired output matches core

        # Pathname prefixes
        TEMP => "",           # Pathname - temp folder prefix
        PARM => "",           # Pathname - parameter folder prefix
        ATRIA => "",          # Pathname - atria folder prefix
        @_,                   # Override previous attributes
        @_,                   # Extra time for ccperl
        };

    bless($self, $class);

    # All calls to create an IOBase class should call its Initialize method
    # once it is created. IOstd() does this automatically.

    if (&GetParm("CQCC_CTRL_C") eq "TRUE") {
        $SIG{INT} = \&IOInterruptHandler;
    }
    return $self;
}

######################################################################
#::method SelfTest

=over 6

=item C<SelfTest(@args)>

This method runs basic unit testing on the object instance which 
currently includes debug print, platform pathname support, prompt
and dialog box support, and file read/write operations. 

=back

=cut

######################################################################

sub SelfTest {
    my ($self, %args) = @_;

    my $basicTest = 1;
    my $basicFileTest = 1;
    my $threadTest = 0;
    my $resMgrTest = 0;
    my $helpDisplayTest = 0;
    my $threadID = $$;
    my $threadCount = 1;
    my $lockOpTest = 0;
    
    for (keys %args) { 
        my $key = $_;
        if ($key eq "threadTest") {
            $basicTest = $basicFileTest = 0;
            $threadTest = 1;
            $threadCount = $args{$key};
        } elsif ($key eq "lockOpTest") {
            $lockOpTest = $args{$key};
        } elsif ($key eq "resMgrTest") {
            $basicTest = $basicFileTest = 0;
            $resMgrTest = 1;
        } elsif ($key eq "helpDisplay") {
            $basicTest = $basicFileTest = 0;
            $helpDisplayTest = 1;
        }
    }
    if ($basicTest)     { 
        $self->basicTest(); 
    }
    if ($basicFileTest) { 
        $self->basicFileTest("write", $threadID, -1); 
        $self->basicFileTest("copy", $threadID, -1); 
        $self->basicFileTest("rename", $threadID, -1); 
    }
    if ($helpDisplayTest) { 
        print "Starting help display\n";
        my $path = $self->Pathname("ATRIA", "../doc/help/cc_cq_int/index.htm");
        $self->HelpDisplay($path);
        print "Launched help display\n";
    }
    if ($threadTest)    { 
        my $i;
        my $result = 1;
        my $lockname = ($lockOpTest && $lockOpTest ne 0) ? 
                       &IOPathname("PARM","lockOpTest.txt") : undef;

        if ($lockname) {
            $self->LockOperation($lockname,$IO_LOCKOP_NET);
        }
        for ($i = 0; $i < $threadCount; $i++) {
            if (! $self->basicFileTest("thread", $threadID, $i)) {
                $result = 0;
            }
            # Release/reacquire lock every 5 steps thru
            if ($lockname && ($i % 5 eq 0)) {
                $self->LockOperation($lockname,$IO_LOCKOP_UNLOCK);
                sleep(1);
                $self->LockOperation($lockname,$IO_LOCKOP_NET);
            }
        }
        if ($lockname) {
            $self->LockOperation($lockname,$IO_LOCKOP_UNLOCK);
        }
        if ($result ne 1) {
            print "ERROR: threadTest failed\n";
        } 
    }
    if ($resMgrTest) {
        my $resource = $self->ResourceManager("testResMgr", 1, 0);
        &IOFileWrite($resource, "[$threadID] Acquired");
        $resource = $self->ResourceManager("testResMgr", 0, 0);

        $resource = $self->ResourceManager("testResMgr", 1, 0);
        &IOFileAppend($resource, "[$threadID] reacquired");
        $resource = $self->ResourceManager(undef, 0, 0);
    }
    return 0;
}

######################################################################
#:::basicTest

sub basicTest {
    my ($self) = @_;
    my $name = "testTrigger";

    ##############################
    print "\nClassTest: print out DebugPrint Test message\n";
    &IODebugPrint("BASIC", "Test message\n");

    ##############################
    print "\nClassTest: print sample temp, parm, and atria paths\n";

    print "Pathname test: temp/foo.c ", 
        &IOPathname("TEMP","foo.c"), "\n";
    print "Pathname test: params/foo.c ", 
        &IOPathname("PARM","foo.c"), "\n";
    print "Pathname test: param_user/foo.c ", 
        &IOPathname("PARM_USER","foo.c"), "\n";
    print "Pathname test: param/foo.SERIES_ID", 
        &IOPathname("PARM_SERIES","foo"), "\n";
    print "Pathname test: atria/bin/foo.c ", 
        &IOPathname("ATRIA","bin/foo"), "\n";

    ##############################
    print "\nClassTest: print out usage message with provided name\n";
    &IOMessage("ERROR", [ "TBASE_I_USAGE", $name, $name, $name ]);
    &IOMessage("INFO", [ "MENU_ASSOCIATION" ]);

    print "\nClassTest: Prompt user for bug list \n";



    my ($abort,$answer) = 
       &IOPrompt(["TCQCC_P_ENTER_BUGS_NOWARN_REQD", "checkout", "<filename>"]);

    print "   Abort: $abort, Answer: $answer\n";

    ##############################
    my @options = ("123 Big Bad Bug", "234 Medium", "345 Little Bug");
    my @choices;
    ($abort, @choices) = &IOListSelect(1, 
         ["TCQCC_P_QUERY_RESULTS","<user>"], @options);

    print "IOListSelect abort: $abort answers: " . 
           join(", ", @choices) . "\n\n";

    ##############################
    @options = ("Top/First/choice1", "Top/First/choice2", 
                "Top/First/choice3", "Top/Second/choice1",
                "Top/Second/choice2","Top2/Third/choice1",
                "Top2/Third/choice2", "Top2/Third/Fourth/choice3");
    ($abort, @choices) = &IOListTreeSelect(1, "/",
             ["TCQCC_P_QUERY_RESULTS","<user>"], @options);

    print "IOListTreeSelect abort: $abort answers: " . 
           join(", ", @choices) . "\n\n";
}

######################################################################
#:::basicFileTest

sub basicFileTest {
    my ($self, $mode, $threadID, $iteration) = @_;
    my $suffix = "[$threadID,$mode,$iteration]";

    ##############################
    my $fname = &IOPathname("PARM","basicFileTest.txt");
    my $readFile = $fname;
    my $user = "My Name";
    my $pass = "mypassword";
    my @list = (1,2,3);
    my $listString = join(' ',@list);

    ##############################
    # Append to file twice, writing two lines each time

    if ($mode ne "thread") {
        &IOFileWrite($fname, $suffix . $user, 
                             $suffix . $pass);
    } else {
        &IOFileAppend($fname, $suffix . $user, 
                              $suffix . $pass);
    }
    &IOFileAppend($fname, $suffix . "Append", 
                          $suffix . $listString);

    if ($mode eq "copy") {
        $readFile = $fname . ".copy";
        $self->FileOperation($IO_FILE_COPY, "FATAL", "basicFileTest", 
                             $fname, $readFile);
    } elsif ($mode eq "rename") {
        $readFile = $fname . ".rename";
        $self->FileOperation($IO_FILE_RENAME, "FATAL", "basicFileTest", 
                             $fname, $readFile);
    }

    my (@lines) = &IOFileRead($readFile, ($mode ne "thread" ? 1 : undef));
    if ($mode eq "copy") { 
        unlink $fname;
    }

    ##############################
    # Read back to ensure all data preserved in order

    my $state = 0;
    for (@lines) {
        my $line = $_;
        if ($line =~ /$suffix/) {
            if ($state eq 0 && $line =~ /$user/) {
                print "Found $suffix$user\n";
                $state++;
            } elsif ($state eq 1 && $line =~ /$pass/) {
                print "Found $suffix$pass\n";
                $state++;
            } elsif ($state eq 2 && $line =~ /Append/) {
                print "Found $suffix" . "Append\n";
                $state++;
            } elsif ($state eq 3 && $line =~ /$listString/) {
                print "Found $suffix$listString\n";
                $state++;
            }
        }
    }
    if ($state ne 4) {
        print "basicFileTest ($mode) $suffix failed, final state $state != 4\n";
        return 0;
    } else {
        print "basicFileTest ($mode) $suffix passed\n";
    }
    return 1;
}

######################################################################
#::method Shutdown

=over 6

=item C<Shutdown(status)>

This method is called by the Exit Handler after the cleanup handler
and other exit operations have been called but before actually exiting.
It can be overridden by subclasses which need to shutdown more completely.

=back

=cut

######################################################################

sub Shutdown {
    my ($self,$status) = @_;
}

######################################################################
#::method _LogEntry

=over 6

=item C<_LogEntry(caller, msgtag, noExitFlag)>

Depending on the current mode, this method will either record a log
entry noting which IO routine is being called or playback the log, 
confirming that the caller/msgtag entry is the next one expected. 

If the noExitFlag is "1", the routine will return 0 if the entry
is not what is expected instead of exiting with a Fatal error, 
leaving it up to the caller to exit instead. 

=back

=cut

######################################################################

sub _LogEntry {
    my ($self, $caller, $msgTag, $noExitFlag) = @_;
    if ($self->{JOURNAL_MODE} eq "OFF") { 
       return;
    } 
    my @msgTagList;
    if (ref($msgTag) eq "ARRAY") { 
       @msgTagList = @{ $msgTag };
    } else {
       push @msgTagList, $msgTag;
    }
    my $logMessage = "<<$caller>> [ " . 
        join(", ", @msgTagList) . "]";
    if ($self->{JOURNAL_MODE} eq "RECORD") {
        $self->Print($logMessage . "\n");
    } elsif ($self->{JOURNAL_MODE} eq "PLAYBACK") { 
        my $line = $self->_LogRead();
        # Read thru log til next <<>> entry, then expect to match that
        # Compare input to $logMessage, error if not identical

        if ($logMessage ne $line) {
            $self->Print("Error: Playback not correct\n " . 
                "Actual: $logMessage\nExpected: $line\n");
            if ($noExitFlag eq "1") { 
                return 0;
            }
            &IOExitHandler(2);
        }
        $self->Print($logMessage . "\n");
    }
    return 1;
}

######################################################################
#::method _LogRead

=over 6

=item C<_LogRead()>

When playback is enabled this method reads the log file up until 
it gets to the next journaling tag <<anything>> and returns that line.

=back

=cut

######################################################################

sub _LogRead {
    my ($self) = @_;
    my $line;

    if ($self->{JOURNAL_MODE} ne "PLAYBACK") { 
       return;
    } 
    if (! $journalHandle) {
        my $fileName = &GetParm("CQCC_JOURNAL_FILE");
        open(JOURNAL_HANDLE, $fileName);
        $journalHandle = 1;
        &IODebugPrint("BASIC", "IOBase::_LogRead " . 
                      "Opening journal file $fileName\n");
    } 
    while ($line = <JOURNAL_HANDLE>) {
        if (substr($line, 0, 2) eq "<<") {
            if ($line =~ /\[/) {
                if (! ($line =~/\]/)) {
                    my $buffer;
                    my $keepGoing=1;
                    while ($keepGoing && ($buffer = <JOURNAL_HANDLE>)) {
                        $line = $line . $buffer;
                        if ($buffer =~ /\]/) {
                            $keepGoing = 0;
                        }
                    }
                }
            }          
            chomp $line;
            return $line;
        }
    }
    &IODebugPrint("BASIC", "IOBase::_LogRead closing journal file\n");

    close(JOURNAL_HANDLE);
    $journalHandle = 0; 
    return "";
}

######################################################################
#::method _ReadSTDIN

=over 6

=item C<_ReadSTDIN(sequence, promptMsgTag, {user_answer})>

Depending on the current mode, this method will either prompt the user
for new input or get input from a playback log, then record the input
to a log if in recording mode. 

The sequence argument is used to differentiate between a single-call 
approach ("DURING") when reading from STDIN and a two-call approach
when a GUI widget is active: "BEFORE" checks for an available answer
from the journal log that would avoid the need to display the widget,
and "AFTER" is called after the user has provided an answer to the 
widget that should be recorded to the journal if we are recording.
In the latter case, the optional user_answer is the widget's user
input that needs to be written to the log. 

=back

=cut

######################################################################

sub _ReadSTDIN {
    my ($self, $sequence, $msgTag, $user_answer) = @_;
    my $answer;
    my $msg = $self->{MSGCAT}->Message( $msgTag );

    #BEFORE/DURING: Confirm next log tag entry is correct
    if ($sequence eq "DURING" || $sequence eq "BEFORE") {
        $self->_LogEntry("_ReadSTDIN", $msgTag);
    }

    #BEFORE/DURING Playback handling
    if (($sequence eq "DURING" || $sequence eq "BEFORE") && 
        ($self->{JOURNAL_MODE} eq "PLAYBACK")) {
        # Read thru log til next <<>> entry, then expect to match that
        # get $answer from log
    
        my $line = $self->_LogRead();
        if ($line =~ m/^<<user>>(.+)$/) { 
            $answer = $1;
        } elsif ($line =~ m/^<<user>>$/) {
            $answer = "";
        } else { 
            &IOPrint("Error: Playback not correct\n " .
                "Expected: <<user>>answer\n\n" . 
                "Got: $line\n");
            IOExitHandler(2);
        }
        &IOPrint("$msg\n<<user>>$answer\n\n");

        # Check <<user>> answers for environment variables 
        my @preitems =split(",",$answer);
        my @postitems;
        my $item;            
        for $item (@preitems) {
            if ($item =~ /ENV:(\S*)/) {
                if ($ENV{$1}) {
                   $item = $ENV{$1};
                } else {
                   $item = "Unknown env var: $item";
                }
            }
            push @postitems, $item;
        }
        $answer = join(",",@postitems);
    }

    #DURING: Prompt the user for actual input and echo it to output
    if ($sequence eq "DURING") {
        if ($self->{JOURNAL_MODE} ne "PLAYBACK") { 
            &IOPrint($msg); 
            if ($ENV{CQCC_NO_STDIN}) {
                &IOMessage("FATAL", ["IO_E_NO_STDIN"]);
                return;
            }
            $answer = <STDIN>;
        }

        if ($self->{JOURNAL_MODE} eq "RECORD") {
            &IOPrint("\n<<user>>$answer\n");
        }
    }

    #AFTER: Just save the provided user_answer as if it was typed in
    if ($sequence eq "AFTER") {
        if ($self->{JOURNAL_MODE} eq "RECORD") { 
            &IOPrint("\n<<user>>$user_answer\n\n");
        }
    }

    chomp $answer;
    return $answer;
}

######################################################################
#::Tail
######################################################################

1;
